# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""PDFO optimization library wrapper, see `PDFO website <https://www.pdfo.net/>`_."""
from __future__ import division

import logging

from numpy import inf, isfinite, real
from pdfo import pdfo

from gemseo.algos.opt.opt_lib import OptimizationLibrary
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class PDFOOpt(OptimizationLibrary):
    """PDFO optimization library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = False

    OPTIONS_MAP = {
        OptimizationLibrary.MAX_ITER: "max_iter",
    }

    def __init__(self):
        """Constructor.

        Generate the library dict, contains the list
        of algorithms with their characteristics:

        - does it require gradient
        - does it handle equality constraints
        - does it handle inequality constraints
        """
        super(PDFOOpt, self).__init__()
        doc = "https://www.pdfo.net/"
        self.lib_dict = {
            "PDFO_COBYLA": {
                self.INTERNAL_NAME: "cobyla",
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: True,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "Constrained Optimization"
                "By Linear Approximations ",
                self.WEBSITE: doc,
            },
            "PDFO_BOBYQA": {
                self.INTERNAL_NAME: "bobyqa",
                self.REQUIRE_GRAD: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Bound Optimization By " "Quadratic Approximation",
                self.WEBSITE: doc,
            },
            "PDFO_NEWUOA": {
                self.INTERNAL_NAME: "newuoa",
                self.REQUIRE_GRAD: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "NEWUOA",
                self.WEBSITE: doc,
            },
        }
        self.name = "PDFO"

    def _get_options(
        self,
        ftol_rel=1e-12,
        ftol_abs=1e-12,
        xtol_rel=1e-12,
        xtol_abs=1e-12,
        max_time=0,
        rhobeg=0.5,
        rhoend=1e-6,
        max_iter=500,
        ftarget=-inf,
        scale=False,
        quiet=True,
        classical=False,
        debug=False,
        chkfunval=False,
        ensure_bounds=True,
        normalize_design_space=True,
        **kwargs
    ):
        r"""Sets the options default values

        To get the best and up to date information about algorithms options,
        go to pdfo documentation on the `PDFO website <https://www.pdfo.net/>`_.

        :param ftol_rel: stop criteria, relative tolerance on the
               objective function,
               if abs(f(xk)-f(xk+1))/abs(f(xk))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_rel: float
        :param ftol_abs: stop criteria, absolute tolerance on the objective
               function, if abs(f(xk)-f(xk+1))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_abs: float
        :param xtol_rel: stop criteria, relative tolerance on the
               design variables,
               if norm(xk-xk+1)/norm(xk)<= xtol_rel: stop
               (Default value = 1e-9)
        :type xtol_rel: float
        :param xtol_abs: stop criteria, absolute tolerance on the
               design variables,
               if norm(xk-xk+1)<= xtol_abs: stop
               (Default value = 1e-9)
        :type xtol_abs: float
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param rhobeg: Initial value of the trust region radius.
        :type max_iter: float
        :param rhoend:  Final value of the trust region radius.  Indicate
        the accuracy required in the final values of the variables
        :type rhoend: float
        :param maxfev:  Upper bound of the number of calls of the objective
        function `fun`.
        :type maxfev: int
        :param ftarget: Target value of the objective function. If a feasible
        iterate achieves an objective function value lower or equal to
        `options['ftarget']`, the algorithm stops immediately.
        :type ftarget: float
        :param scale: Flag indicating whether to scale the problem according to
        the bound constraints.
        :type scale: bool
        :param quiet: Flag of quietness of the interface. If it is set to True,
        the output message will not be printed.
        :type quiet: bool
        :param classical: Flag indicating whether to call the classical Powell
        code or not.
        :type classical: bool
        :param debug: Debugging flag.
        :type debug: bool
        :param chkfunval: Flag used when debugging. If both `options['debug']`
        and `options['chkfunval']` are True, an extra function/constraint
        evaluation would be performed to check whether the returned values of
        objective function and constraint match the returned x.
        :type chkfunval: bool
        :param ensure_bounds:
        :type ensure_bounds: bool
        :param normalize_design_space: If True, normalize the design space
        :type normalize_design_space: bool
        """
        nds = normalize_design_space
        popts = self._process_options(
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            xtol_abs=xtol_abs,
            max_time=max_time,
            rhobeg=rhobeg,
            rhoend=rhoend,
            max_iter=max_iter,
            ftarget=ftarget,
            scale=scale,
            quiet=quiet,
            classical=classical,
            debug=debug,
            chkfunval=chkfunval,
            ensure_bounds=ensure_bounds,
            normalize_design_space=nds,
            **kwargs
        )
        return popts

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm
        """
        # remove normalization from options for algo
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)

        # Get the normalized bounds:
        x_0, l_b, u_b = self.get_x0_and_bounds_vects(normalize_ds)

        # Ensure bounds
        ensure_bounds = options["ensure_bounds"]

        # Replace infinite values with None:
        l_b = [val if isfinite(val) else None for val in l_b]
        u_b = [val if isfinite(val) else None for val in u_b]
        bounds = list(zip(l_b, u_b))

        def real_part_fun(x_vect):
            """Wraps the function and returns the real part."""
            return real(self.problem.objective.func(x_vect))

        if ensure_bounds:
            fun = self.ensure_bounds(real_part_fun, normalize_ds)
        else:
            fun = real_part_fun

        constraints = self.get_right_sign_constraints()

        cstr_scipy = []
        for cstr in constraints:
            if PY2:
                f_type = cstr.f_type.encode("ascii")
            else:
                f_type = cstr.f_type
            if ensure_bounds:
                c_scipy = {
                    "type": f_type,
                    "fun": self.ensure_bounds(cstr.func, normalize_ds),
                }
            else:
                c_scipy = {"type": f_type, "fun": cstr.func}

            cstr_scipy.append(c_scipy)

        # |g| is in charge of ensuring max iterations, since it may
        # have a different definition of iterations, such as for SLSQP
        # for instance which counts duplicate calls to x as a new iteration
        max_iter = options[self.MAX_ITER]
        options["maxfev"] = int(max_iter * 1.2)

        opt_result = pdfo(
            fun=fun,
            x0=x_0,
            method=self.internal_algo_name,
            bounds=bounds,
            constraints=cstr_scipy,
            options=options,
        )

        return self.get_optimum_from_database(opt_result.message, opt_result.status)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Scalable module from Tedford and Martins (2010)
***********************************************

The modules located in the **scalable_tm** directory offer a set of
classes relative to the scalable problem introduced in the paper:

    Tedford NP, Martins JRRA (2010), Benchmarking
    multidisciplinary design optimization algorithms,
    Optimization and Engineering, 11(1):159-183.

Overview
~~~~~~~~

This scalable problem aims to minimize an objective function
quadratically depending on shared design parameters and coupling variables,
under inequality constraints linearly depending on these coupling variables.

System discipline
-----------------

A system discipline computes the constraints and the objective
in function of the shared design parameters and coupling variables.

Strongly coupled disciplines
----------------------------

The coupling variables are the outputs of strongly coupled disciplines.

Each strongly coupled discipline computes a set of coupling variables
linearly depending on local design parameters, shared design parameters,
coupling variables from other strongly coupled disciplines,
and belonging to the unit hypercube.

Scalability
-----------

This problem is said "scalable"
because several sizing features can be chosen by the user:

- the number of local design parameters for each discipline,
- the number of shared design parameters,
- the number of coupling variables for each discipline,
- the number of disciplines.

A given sizing configuration is called "scaling strategy"
and this scalable module is particularly useful to compare different MDO
formulations with respect to the scaling strategy.

Implementation
~~~~~~~~~~~~~~

The scalable problem
--------------------

The :class:`.TMScalableProblem` class instantiates the disciplines of the
problem, that are :class:`.TMMainDiscipline` and several :class:`.TMSubDiscipline`,
as well as the  :class:`.DesignSpace`.
These instantiated objects can be used in a :class:`.Scenario`,
e.g. :class:`.MDOScenario` or :class:`.DOEScenario`.

The scalable study
------------------

The :class:`.TMScalableStudy` class instantiates a :class:`.TMScalableProblem`
for a particular scaling strategy
where the number of local design parameters is the same for all disciplines
as well as the number of coupling variables.
It provides a method to run MDO formulations and graphical capabilities
to analyze the results.

The parametric scalable study
-----------------------------

The :class:`.TMParamSS` class instantiates several :class:`.TMScalableStudy`
associated with different scaling strategies, e.g. different numbers of local
design parameters or different numbers of coupling variables.
It provides a method to run MDO formulations and save results on the disk.
The :class:`.TMParamSSPost` provides graphical capabilities to post-process
these saved results.
"""



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Correlations in the optimization database."""

from __future__ import division, unicode_literals

import logging
from typing import List, Optional, Sequence, Tuple

import matplotlib.gridspec as gridspec
import numpy as np
import pylab
from matplotlib import ticker
from matplotlib.figure import Figure
from numpy import atleast_2d, ndarray

from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class Correlations(OptPostProcessor):
    """Scatter plots of the correlated variables.

    These variables can be design variables, outputs functions or constraints.

    The plot method considers all the variable correlations greater than 95%.
    Another level value, a sublist of variable names or both can be passed as options.
    The x- and y-figure sizes can also be modified.
    """

    def _plot(
        self,
        func_names=None,  # type: Optional[Sequence[str]]
        coeff_limit=0.95,  # type: float
        n_plots_x=5,  # type: int
        n_plots_y=5,  # type: int
    ):  # type: (...) -> None
        """
        Args:
            func_names: The function names subset
                for which the correlations are computed.
                If None, all functions are considered.
            coeff_limit: The plot is not made
                if the correlation between the variables is lower than this limit.
            n_plots_x: The number of horizontal plots.
            n_plots_y: The number of vertical plots.

        Raises:
            ValueError: If an element of `func_names` is not a function
                defined in `opt_problem`.
        """
        functions = self.opt_problem.get_all_functions()
        all_func_names = [func.name for func in functions]

        if not func_names:
            func_names = all_func_names
        elif set(func_names).issubset(all_func_names):
            func_names = [
                func_name for func_name in all_func_names if func_name in func_names
            ]
        else:
            raise ValueError(
                "The following elements are not "
                "functions: {}. Defined functions are {}.".format(
                    ", ".join(set(func_names) - set(all_func_names)),
                    ", ".join(all_func_names),
                )
            )

        values_array, variables_names, _ = self.database.get_history_array(
            func_names, None, True, 0.0
        )

        variables_names = self.__sort_variables_names(variables_names, func_names)

        corr_coeffs_array = self.__compute_correlations(values_array)
        i_corr, j_corr = np.where(
            (np.abs(corr_coeffs_array) > coeff_limit)
            & (np.abs(corr_coeffs_array) < (1.0 - 1e-9))
        )
        LOGGER.info("Detected %s correlations > %s", i_corr.size, coeff_limit)

        if i_corr.size <= 16:
            n_plots_x = 4
            n_plots_y = 4
        spec = gridspec.GridSpec(n_plots_y, n_plots_x, wspace=0.3, hspace=0.75)
        spec.update(top=0.95, bottom=0.06, left=0.08, right=0.95)
        fig = None
        fig_indx = 0
        for plot_index, (i, j) in enumerate(zip(i_corr, j_corr)):
            plot_index_loc = plot_index % (n_plots_x * n_plots_y)
            if plot_index_loc == 0:
                if fig is not None:  # Save previous plot
                    fig_indx += 1
                    self._add_figure(fig)
                fig = pylab.plt.figure()
                mng = pylab.plt.get_current_fig_manager()
                mng.resize(1200, 900)
                ticker.MaxNLocator(nbins=3)

            self.__create_sub_correlation_plot(
                i,
                j,
                corr_coeffs_array[i, j],
                fig,
                spec,
                plot_index_loc,
                n_plots_y,
                n_plots_x,
                values_array,
                variables_names,
            )
        if fig is not None:
            self._add_figure(fig)

    def __create_sub_correlation_plot(
        self,
        i_ind,  # type: int
        j_ind,  # type: int
        corr_coeff,  # type: ndarray
        fig,  # type: Figure
        spec,  # type: gridspec
        plot_index,  # type: int
        n_plot_v,  # type: int
        n_plot_h,  # type: int
        values_array,  # type: ndarray
        variables_names,  # type: Sequence[str]
    ):  # type: (...)-> None
        """Create a correlation plot.

        Args:
            i_ind: The index for the x-axis data.
            j_ind: The index for the y-axis data.
            corr_coeff: The correlation coefficients.
            fig: The figure where the subplot will be placed.
            spec: The matplotlib grid structure.
            plot_index: The local plot index.
            n_plot_v: The number of vertical plots.
            n_plot_h: The number of horizontal plots.
            values_array: The function values from the optimization history.
            variables_names: The variables names.
        """
        gs_curr = spec[int(plot_index / n_plot_v), plot_index % n_plot_h]
        ax1 = fig.add_subplot(gs_curr)
        x_plt = values_array[:, i_ind]
        y_plt = values_array[:, j_ind]
        ax1.scatter(x_plt, y_plt, c="b", s=30)
        self.out_data_dict[(i_ind, j_ind)] = (
            variables_names[i_ind],
            variables_names[j_ind],
            corr_coeff,
        )
        ax1.set_xlabel(variables_names[i_ind], fontsize=9)
        # Update y labels spacing
        start, stop = ax1.get_ylim()
        ax1.yaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
        start, stop = ax1.get_xlim()
        ax1.xaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
        ax1.set_ylabel(variables_names[j_ind], fontsize=10)
        ax1.tick_params(labelsize=10)
        ax1.set_title("R={:.5f}".format(corr_coeff), fontsize=12)
        ax1.grid()

    @classmethod
    def __compute_correlations(
        cls, values_array  # type: ndarray
    ):  # type: (...)-> ndarray
        """Compute correlations.

        Args:
            values_array: The values to compute the correlations.

        Returns:
            The lower diagonal of the correlations matrix.
        """
        ccoeff = np.corrcoef(values_array.astype(float), rowvar=False)
        return np.tril(atleast_2d(ccoeff))  # Keep lower diagonal only

    def __sort_variables_names(
        self,
        variables_names,  # type: Sequence[str]
        func_names,  # type: Sequence[str]
    ):  # type: (...)-> List[str]
        """Sort the expanded variable names using func_names as the pattern.

        In addition to sorting the expanded variable names, this method
        replaces the default hard-coded vectors (x_1, x_2, ... x_n) with
        the names given by the user.

        Args:
            variables_names: The expanded variable names to be sorted.
            func_names: The functions names in the required order.

        Returns:
            The sorted expanded variable names.
        """

        def func_order(
            x,  # type: str
        ):  # type: (...) -> Tuple[int, str]
            """Key function to sort function components.

            Args:
                x: An element from a list.

            Returns:
                The Tuple to use in the sort method.
            """

            for i, func_name in enumerate(func_names):
                if x.find(func_name) == 0:
                    return (i, x.replace(func_name, ""))

            return (len(func_names) + 1, x)

        variables_names.sort(key=func_order)
        x_names = self._generate_x_names()
        return variables_names[: -len(x_names)] + x_names



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A Pareto Front."""
from __future__ import division, unicode_literals

import logging
from typing import List, Optional, Sequence, Tuple

from numpy import full, ndarray

from gemseo.algos.pareto_front import generate_pareto_plots
from gemseo.post.opt_post_processor import OptPostProcessor

LOGGER = logging.getLogger(__name__)


class ParetoFront(OptPostProcessor):
    """Compute the Pareto front Search for all non dominated points.

    For each point, check if it exists ``j`` such that there is no lower value
    for ``obj_values[:,j]`` that does not degrade
    at least one other objective ``obj_values[:,i]``.

    Generate a plot or a matrix of plots if there are more than 2 objectives.
    Plot in red the locally non dominated points for the currrent two objectives.
    Plot in green the globally (all objectives) Pareto optimal points.
    """

    def _plot(
        self,
        objectives=None,  # type: Optional[Sequence[str]]
        objectives_labels=None,  # type: Optional[Sequence[str]]
        figsize_x=10.0,  # type: float
        figsize_y=10.0,  # type: float
        show_non_feasible=True,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            objectives: The functions names or design variables to plot.
                If None, use the objective function (may be a vector).
            objectives_labels: The labels of the objective components.
                If None, use the objective name suffixed by an index.
            figsize_x: The size of figure in the horizontal direction (inches).
            figsize_y: The size of figure in the vertical direction (inches).
            show_non_feasible: If True, show the non feasible points in the plot.

        Raises:
            ValueError: If the numbers of objectives and objectives
             labels are different.
        """
        if objectives is None:
            objectives = [self.opt_problem.objective.name]

        all_funcs = self.opt_problem.get_all_functions_names()
        all_dv_names = self.opt_problem.design_space.variables_names

        vals, vname = self.__compute_names_and_values(
            all_dv_names, all_funcs, objectives
        )

        non_feasible_samples = self.__compute_non_feasible_samples(vals)

        if objectives_labels is not None:
            if len(vname) != len(objectives_labels):
                raise ValueError(
                    "objective_labels shall have the same dimension as vname."
                )
            vname = objectives_labels

        fig = generate_pareto_plots(
            vals,
            vname,
            figsize=(figsize_x, figsize_y),
            non_feasible_samples=non_feasible_samples,
            show_non_feasible=show_non_feasible,
        )

        self._add_figure(fig)

    def __compute_names_and_values(
        self,
        all_dv_names,  # type: Sequence[str]
        all_funcs,  # type: Sequence[str]
        objectives,  # type: Sequence[str]
    ):  # type: (...) -> Tuple[ndarray,List[str]]
        """Compute the names and values of the objective and design variables.

        Args:
             add_dv_names: The design variables names.
             all_funcs: The function names.
             objectives: The objective names.

        Returns:
            The sample values and the sample names.
        """
        # TODO: Those lines are not covered by the tests. It has to be
        # investigated to see if it's dead code.
        if not objectives:
            # The function list only contains design variables
            vals = self.database.get_x_history()
            vname = self.database.set_dv_names(vals[0].shape[0])
        else:
            design_variables = []
            for func in list(objectives):
                self.__check_objective_name(all_dv_names, all_funcs, func, objectives)
                self.__move_objective_to_design_variable(
                    design_variables, func, objectives
                )

            if not design_variables:
                design_variables = None
                add_dv = False
            else:
                add_dv = True

            vals, vname, _ = self.database.get_history_array(
                objectives, design_variables, add_dv=add_dv
            )
        return vals, vname

    def __check_objective_name(
        self,
        all_dv_names,  # type: Sequence[str]
        all_funcs,  # type: Sequence[str]
        func,  # type: str
        objectives,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Check that the objective name is valid.

        Args:
             add_dv_names: The design variables names.
             all_funcs: The function names.
             func: The function name.
             objectives: The objectives names.

        Raises:
            ValueError: If the objective name is not valid.
        """
        if func not in all_funcs and func not in all_dv_names:
            min_f = "-" + func == self.opt_problem.objective.name
            if min_f and not self.opt_problem.minimize_objective:
                objectives[objectives.index(func)] = "-" + func
            else:
                msg = (
                    "Cannot build Pareto front,"
                    " Function {} is neither among"
                    " optimization problem functions: "
                    "{} nor design variables: {}."
                )
                msg = msg.format(func, str(all_funcs), str(all_dv_names))
                raise ValueError(msg)

    def __move_objective_to_design_variable(
        self,
        design_variables,  # type: Sequence[str]
        func,  # type: str
        objectives,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Move an objective to a design variable.

        If the given function is a design variable,
        then move it from the objectives to the design_variables.

        Args:
             design_variables: The design variables.
             func: The function name.
             objectives: The objectives names.
        """
        if func in self.opt_problem.design_space.variables_names:
            objectives.remove(func)
            design_variables.append(func)

    def __compute_non_feasible_samples(
        self, vals  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the non-feasible indexes.

        Args:
            vals: The sample values.

        Returns:
              An array of size ``n_samples``, True if the point is non feasible
        """
        x_feasible, _ = self.opt_problem.get_feasible_points()
        feasible_indexes = [self.database.get_index_of(x) for x in x_feasible]

        is_non_feasible = full(vals.shape[0], True)
        is_non_feasible[feasible_indexes] = False

        return is_non_feasible



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
An advanced MDA splitting algorithm based on graphs
***************************************************
"""
from __future__ import division, unicode_literals

import logging
from multiprocessing import cpu_count
from os.path import join, split

from gemseo.api import create_mda
from gemseo.core.chain import MDOChain
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import SerialExecSequence
from gemseo.mda.mda import MDA

LOGGER = logging.getLogger(__name__)
N_CPUS = cpu_count()


class MDAChain(MDA):
    """The **MDAChain** computes a chain of subMDAs and simple evaluations.

    The execution sequence is provided by the
    :class:`.DependencyGraph` class.
    """

    def __init__(
        self,
        disciplines,
        sub_mda_class="MDAJacobi",
        max_mda_iter=20,
        name=None,
        n_processes=N_CPUS,
        chain_linearize=False,
        tolerance=1e-6,
        use_lu_fact=False,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        **sub_mda_options
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param sub_mda_class: the class to instantiate for sub MDAs
        :type sub_mda_class: str
        :param max_mda_iter: maximum number of iterations for sub MDAs
        :type max_mda_iter: int
        :param name: name of self
        :type name: str
        :param n_processes: number of processes for parallel run
        :type n_processes: int
        :param chain_linearize: linearize the chain of execution, if True
            Otherwise, linearize the oveall MDA with base class method
            Last option is preferred to minimize computations in adjoint mode
            in direct mode, chain_linearize may be cheaper
        :type chain_linearize: bool
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :param sub_mda_options: options dict passed to the sub mda
        :type sub_mda_options: dict
        """
        self.n_processes = n_processes
        self.mdo_chain = None
        self.__chain_linearize = chain_linearize
        self.sub_mda_list = []

        # compute execution sequence of the disciplines
        super(MDAChain, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            tolerance=tolerance,
            use_lu_fact=use_lu_fact,
            grammar_type=grammar_type,
        )

        if (
            not self.coupling_structure.get_all_couplings()
            and not self.__chain_linearize
        ):
            LOGGER.warning("No coupling in MDA, switching chain_linearize to True")
            self.__chain_linearize = True

        self._create_mdo_chain(
            disciplines, sub_mda_class=sub_mda_class, **sub_mda_options
        )
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()
        # cascade the tolerance
        for sub_mda in self.sub_mda_list:
            sub_mda.tolerance = self.tolerance

    def _create_mdo_chain(
        self, disciplines, sub_mda_class="MDAJacobi", **sub_mda_options
    ):
        """Create an MDO chain from the execution sequence of the disciplines.

        :param sub_mda_class: class of sub-MDAs in
            {Jacobi, GS, Newton, Sequential}
        :param disciplines: list of disciplines
        :param sub_mda_options: options passed to the MDA at construction
        """
        chained_disciplines = []
        self.sub_mda_list = []

        for parallel_tasks in self.coupling_structure.sequence:
            # to parallelize, check if 1 < len(parallel_tasks)
            # for now, parallel tasks are run sequentially
            for coupled_disciplines in parallel_tasks:
                first_disc = coupled_disciplines[0]
                if len(coupled_disciplines) > 1 or (
                    len(coupled_disciplines) == 1
                    and self.coupling_structure.is_self_coupled(first_disc)
                ):
                    # several disciplines coupled

                    # order the MDA disciplines the same way as the
                    # original disciplines
                    sub_mda_disciplines = []
                    for disc in disciplines:
                        if disc in coupled_disciplines:
                            sub_mda_disciplines.append(disc)

                    # create a sub-MDA
                    sub_mda = create_mda(
                        sub_mda_class,
                        sub_mda_disciplines,
                        max_mda_iter=self.max_mda_iter,
                        tolerance=self.tolerance,
                        grammar_type=self.grammar_type,
                        **sub_mda_options
                    )
                    sub_mda.n_processes = self.n_processes

                    chained_disciplines.append(sub_mda)
                    self.sub_mda_list.append(sub_mda)
                else:
                    # single discipline
                    chained_disciplines.append(first_disc)

        # create the MDO chain that sequentially evaluates the sub-MDAs and the
        # single disciplines
        self.mdo_chain = MDOChain(
            chained_disciplines, name="MDA chain", grammar_type=self.grammar_type
        )

    def _initialize_grammars(self):
        """Define all inputs and outputs of the chain."""
        self.input_grammar.update_from(self.mdo_chain.input_grammar)
        self.output_grammar.update_from(self.mdo_chain.output_grammar)

    def _run(self):
        """Execute the chained MDA."""
        if self.warm_start:
            self._couplings_warm_start()
        self.local_data = self.mdo_chain.execute(self.local_data)
        return self.local_data

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Actual computation of the jacobians.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization
            should be performed wrt all inputs
            (Default value = None)
        :param outputs: linearization should be performed on
            outputs list.
            If None, linearization should be
            performed on all chain_outputs (Default value = None)
        """
        if self.__chain_linearize:
            self.mdo_chain.add_differentiated_inputs(inputs)
            self.mdo_chain.add_differentiated_outputs(outputs)
            # the Jacobian of the MDA chain is the Jacobian of the MDO chain
            last_cached = self.cache.get_last_cached_inputs()
            self.mdo_chain.linearize(last_cached)
            self.jac = self.mdo_chain.jac
        else:
            super(MDAChain, self)._compute_jacobian(inputs, outputs)

    def add_differentiated_inputs(self, inputs=None):
        """Add inputs to the differentiation list.

        Updates self._differentiated_inputs with inputs

        :param inputs: list of inputs variables to differentiate
            if None, all inputs of discipline are used (Default value = None)
        """
        MDA.add_differentiated_inputs(self, inputs)
        if self.__chain_linearize:
            self.mdo_chain.add_differentiated_inputs(inputs)

    def add_differentiated_outputs(self, outputs=None):
        """Add outputs to the differentiation list.

        Updates self._differentiated_inputs with inputs

        :param outputs: list of output variables to differentiate
            if None, all outputs of discipline are used
        """
        MDA.add_differentiated_outputs(self, outputs=outputs)
        if self.__chain_linearize:
            self.mdo_chain.add_differentiated_outputs(outputs)

    @property
    def normed_residual(self):
        """Accessor to the normed_residuals, computed from the sub_mdas residuals."""
        return sum((mda.normed_residual ** 2 for mda in self.sub_mda_list)) ** 0.5

    @normed_residual.setter
    def normed_residual(self, normed_residual):
        """Set the normed_residual.

        Has no effect, since the normed residuals are defined by sub mdas residuals
        (see associated property)

        Here for compatibility with mother class.
        """
        pass

    def get_expected_dataflow(self):
        """Get the expected dataflow.

        See MDOChain.get_expected_dataflow
        """
        return self.mdo_chain.get_expected_dataflow()

    def get_expected_workflow(self):
        """Get the expected workflow.

        See MDOChain.get_expected_workflow
        """
        exec_s = SerialExecSequence(self)
        workflow = self.mdo_chain.get_expected_workflow()
        exec_s.extend(workflow)
        return exec_s

    def reset_statuses_for_run(self):
        """Set all the statuses to PENDING."""
        super(MDAChain, self).reset_statuses_for_run()
        self.mdo_chain.reset_statuses_for_run()

    def plot_residual_history(
        self,
        show=False,
        save=True,
        n_iterations=None,
        logscale=None,
        filename=None,
        figsize=(50, 10),
    ):
        """Generate a plot of the residual history All residuals are stored in the
        history ; only the final residual of the converged MDA is plotted at each
        optimization iteration.

        :param show: if True, displays the plot on screen
            (Default value = False)
        :param save: if True, saves the plot as a PDF file
            (Default value = True)
        :param n_iterations: if not None, fix the number of iterations in
            the x axis (Default value = None)
        :param logscale: if not None, fix the logscale in the y axis
            (Default value = None)
        :param filename: Default value = None)
        """
        for sub_mda in self.sub_mda_list:
            if filename is not None:
                s_filename = split(filename)
                filename = join(
                    s_filename[0], sub_mda.__class__.__name__ + "_" + s_filename[1]
                )
            sub_mda.plot_residual_history(
                show, save, n_iterations, logscale, filename, figsize
            )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A scatter plot matrix to display optimization history."""
from __future__ import division, unicode_literals

import logging
from typing import Sequence

from matplotlib import pyplot
from pandas.core.frame import DataFrame

from gemseo.post.opt_post_processor import OptPostProcessor

try:
    from pandas.tools.plotting import scatter_matrix
except ImportError:
    from pandas.plotting import scatter_matrix


LOGGER = logging.getLogger(__name__)


class ScatterPlotMatrix(OptPostProcessor):
    """Scatter plot matrix among design variables, output functions and constraints.

    The list of variable names has to be passed as arguments of the plot method.
    x- and y- figure sizes can be changed in option.
    """

    def _plot(
        self,
        variables_list,  # type: Sequence[str]
        figsize_x=10,  # type: int
        figsize_y=10,  # type: int
    ):  # type: (...) -> None
        """
        Args:
            variables_list: The functions names or design variables to plot.
                If the list is empty,
                plot all design variables.
            figsize_x: The size of the figure in horizontal direction (inches).
            figsize_y: The size of the figure in vertical direction (inches).
        """

        add_dv = False
        all_funcs = self.opt_problem.get_all_functions_names()
        all_dv_names = self.opt_problem.design_space.variables_names
        variables_list.sort()

        if not variables_list:
            # In this case, plot all design variables, no functions.
            vals = self.database.get_x_history()

            # This section creates readable labels for design variables
            # i.e. toto_0, toto_1 if toto is a variable with 2 components
            x_labels = self.__get_design_var_labels(all_dv_names)

        else:
            design_variables = []
            for func in list(variables_list):
                if func not in all_funcs and func not in all_dv_names:
                    min_f = "-{}".format(func) == self.opt_problem.objective.name
                    if min_f and not self.opt_problem.minimize_objective:
                        variables_list[variables_list.index(func)] = "-{}".format(func)
                        variables_list.sort()
                    else:
                        msg = (
                            "Cannot build scatter plot matrix, "
                            "Function {} is neither among"
                            " optimization problem functions : {}"
                            " nor design variables : {}".format(
                                func, all_funcs, all_dv_names
                            )
                        )
                        raise ValueError(msg)
                if func in self.opt_problem.design_space.variables_names:
                    # if given function is a design variable, then remove it
                    add_dv = True
                    variables_list.remove(func)
                    design_variables.append(func)
            if not design_variables:
                design_variables = None

            if add_dv:
                # Sort the design variables to be consistent with GEMSEO.
                design_variables = sorted(
                    set(all_dv_names) & set(design_variables),
                    key=all_dv_names.index,
                )

                # This section creates readable labels for design variables
                # and functions i.e. toto_0, toto_1 if toto is a variable
                # with 2 components
                dv_labels = self.__get_design_var_labels(design_variables)
                if variables_list:
                    _, func_labels, _ = self.database.get_history_array(
                        functions=variables_list,
                        design_variables_names=None,
                        add_dv=False,
                    )
                else:
                    func_labels = []
                # vname contains function names + condensed variable names
                # i.e. "toto" even if toto has 2 components or more
                vname = variables_list + design_variables
                # x_labels contains function names + readable variable names
                x_labels = func_labels + dv_labels
            else:
                # In this case we are only plotting functions.
                # Functions have unique names, so x_labels and
                # vname are equal.
                vname = variables_list
                _, x_labels, _ = self.database.get_history_array(
                    functions=variables_list,
                    design_variables_names=None,
                    add_dv=False,
                )
                x_labels.sort()
            dataset = self.opt_problem.export_to_dataset("OptimizationProblem")
            vals = dataset.get_data_by_names(vname, False)
        # Next line is a trick for a bug workaround in numpy/matplotlib
        # https://stackoverflow.com/questions/39180873/
        # pandas-dataframe-valueerror-num-must-be-1-num-0-not-1
        vals = (list(x) for x in vals)
        frame = DataFrame(vals, columns=x_labels)
        scatter_matrix(frame, alpha=1.0, figsize=(figsize_x, figsize_y), diagonal="kde")
        fig = pyplot.gcf()
        fig.tight_layout()
        self._add_figure(fig)

    def __get_design_var_labels(self, des_vars):
        """Create labels for design variables.

        Args:
            des_vars (list(str)): The design variables to get its labels.

        Returns:
            list(str): The labels for the design variables.
        """

        dv_names = []
        for d_v in des_vars:
            dv_size = self.opt_problem.design_space.variables_sizes[d_v]
            if dv_size == 1:
                dv_names.append(d_v)
            else:
                for k in range(dv_size):
                    dv_names.append("{}_{}".format(d_v, k))
        return dv_names



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                  initial documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable model
==============

This module implements the abstract concept of scalable model
which is used by scalable disciplines. A scalable model is built
from a input-output learning dataset associated with a function
and generalizing its behavior to a new user-defined problem dimension,
that is to say new user-defined input and output dimensions.

The concept of scalable model is implemented
through :class:`.ScalableModel`, an abstract class which is instantiated from:

- data provided as a :class:`.Dataset`
- variables sizes provided as a dictionary
  whose keys are the names of inputs and outputs
  and values are their new sizes.
  If a variable is missing, its original size is considered.

Scalable model parameters can also be filled in.
Otherwise the model uses default values.

.. seealso::

   The :class:`.ScalableDiagonalModel` class overloads :class:`.ScalableModel`.
"""
from __future__ import division, unicode_literals

import logging

from numpy import ones, where, zeros

from gemseo.core.dataset import Dataset

LOGGER = logging.getLogger(__name__)


class ScalableModel(object):
    """Scalable model."""

    ABBR = "sm"

    def __init__(self, data, sizes=None, **parameters):
        """Constructor.

        :param Dataset data: learning dataset.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes.
            Default: None.
        :param parameters: model parameters
        """
        sizes = sizes or {}
        self.name = self.ABBR + "_" + data.name
        self.data = data
        self.sizes = self._set_sizes(sizes)
        self.parameters = parameters
        self.lower_bounds, self.upper_bounds = self.compute_bounds()
        self.normalize_data()
        self.lower_bounds, self.upper_bounds = self.compute_bounds()
        self.default_inputs = self._set_default_inputs()
        self.model = self.build_model()

    def _set_default_inputs(self):
        """Sets the default values of inputs from the model.

        :return: default inputs.
        :rtype: dict
        """
        default_inputs = {}
        for name in self.inputs_names:
            default_inputs[name] = 0.5 * ones(self.sizes[name])
        return default_inputs

    def scalable_function(self, input_value=None):
        """Evaluate the scalable function.

        :param dict input_value: input values.
            If None, use default inputs. Default: None.
        :return: evaluation of the scalable function.
        :rtype: dict
        """
        raise NotImplementedError

    def scalable_derivatives(self, input_value=None):
        """Evaluate the scalable derivatives.

        :param dict input_value: input values.
            If None, use default inputs. Default: None
        :return: evaluation of the scalable derivatives.
        :rtype: dict
        """
        raise NotImplementedError

    def compute_bounds(self):
        """Compute lower and upper bounds of both input and output variables.

        :return: lower bounds, upper bounds.
        :rtype: dict, dict
        """
        inputs = self.data.get_data_by_group(self.data.INPUT_GROUP, True)
        outputs = self.data.get_data_by_group(self.data.OUTPUT_GROUP, True)
        lower_bounds = {name: value.min(0) for name, value in inputs.items()}
        lower_bounds.update({name: value.min(0) for name, value in outputs.items()})
        upper_bounds = {name: value.max(0) for name, value in inputs.items()}
        upper_bounds.update({name: value.max(0) for name, value in outputs.items()})

        return lower_bounds, upper_bounds

    def normalize_data(self):
        """Normalize dataset from lower and upper bounds."""
        normalized_data = Dataset()
        inputs = self.data.get_data_by_group(self.data.INPUT_GROUP, True)
        for name in self.data.get_names(self.data.INPUT_GROUP):
            data = inputs[name] - self.lower_bounds[name]
            data /= self.upper_bounds[name] - self.lower_bounds[name]
            normalized_data.add_variable(name, data, self.data.INPUT_GROUP, True)
        outputs = self.data.get_data_by_group(self.data.OUTPUT_GROUP, True)
        for name in self.data.get_names(self.data.OUTPUT_GROUP):
            indices = where(self.lower_bounds[name] == self.upper_bounds[name])[0]
            data = zeros(outputs[name].shape)
            if len(indices) != 0:
                data[:, indices] = zeros((data.shape[0], len(indices))) + 0.5
                self.lower_bounds[name][indices] = zeros(len(indices)) + 0.5
                self.upper_bounds[name][indices] = zeros(len(indices)) + 0.5
            indices = where(self.lower_bounds[name] != self.upper_bounds[name])[0]
            value = outputs[name][:, indices]
            lower_bound = self.lower_bounds[name][indices]
            upper_bound = self.upper_bounds[name][indices]
            data[:, indices] = (value - lower_bound) / (upper_bound - lower_bound)
            normalized_data.add_variable(name, data, self.data.OUTPUT_GROUP, True)
        self.data = normalized_data

    def build_model(self):
        """Build model with original sizes for input and output variables."""
        raise NotImplementedError

    @property
    def original_sizes(self):
        """Original sizes of variables.

        :return: original sizes of variables.
        :rtype: dict
        """
        return self.data.sizes

    @property
    def outputs_names(self):
        """Outputs names.

        :return: names of the outputs.
        :rtype: list(str)
        """
        return sorted(self.data.get_names(self.data.OUTPUT_GROUP))

    @property
    def inputs_names(self):
        """Inputs names.

        :return: names of the inputs.
        :rtype: list(str)
        """
        return sorted(self.data.get_names(self.data.INPUT_GROUP))

    def _set_sizes(self, sizes):
        """Set the new sizes of input and output variables.

        :param sizes: new sizes of some variables.
        :return: new sizes of all variables.
        :rtype: dict
        """
        for group in [self.data.INPUT_GROUP, self.data.OUTPUT_GROUP]:
            for name in self.data.get_names(group):
                original_size = self.original_sizes.get(name)
                sizes[name] = sizes.get(name, original_size)
        return sizes



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A formulation for uncoupled or weakly coupled problems."""
from __future__ import division, unicode_literals

from typing import List, Sequence, Tuple

from gemseo.algos.design_space import DesignSpace
from gemseo.core.chain import MDOChain
from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequence, ExecutionSequenceFactory
from gemseo.core.formulation import MDOFormulation
from gemseo.utils.data_conversion import DataConversion


class DisciplinaryOpt(MDOFormulation):
    """The disciplinary optimization.

    This formulation draws the architecture of a mono-disciplinary optimization process
    from an ordered list of disciplines, an objective function and a design space.
    """

    def __init__(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
        objective_name,  # type: str
        design_space,  # type: DesignSpace
        maximize_objective=False,  # type: bool
    ):  # type: (...) -> None
        self.chain = None
        if len(disciplines) > 1:
            self.chain = MDOChain(disciplines)
        super(DisciplinaryOpt, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        self._filter_design_space()
        self._set_defaultinputs_from_ds()
        # Build the objective from its objective name
        self._build_objective_from_disc(objective_name)

    def get_expected_workflow(
        self,
    ):  # type: (...) -> List[ExecutionSequence,Tuple[ExecutionSequence]]
        if self.chain is None:
            return ExecutionSequenceFactory.serial(self.disciplines[0])
        return self.chain.get_expected_workflow()

    def get_expected_dataflow(
        self,
    ):  # type: (...) -> List[Tuple[MDODiscipline,MDODiscipline,List[str]]]
        if self.chain is None:
            return []
        return self.chain.get_expected_dataflow()

    def get_top_level_disc(self):  # type: (...) -> List[MDODiscipline]
        if self.chain is not None:
            return [self.chain]
        return self.disciplines

    def _filter_design_space(self):  # type: (...) -> None
        """Filter the design space to keep only available variables."""
        all_inpts = DataConversion.get_all_inputs(self.get_top_level_disc())
        kept = set(self.design_space.variables_names) & set(all_inpts)
        self.design_space.filter(kept)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Jean-Christophe Giret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Wrapper of the multiprocessing module in order to circumvent the issues encountered
on Windows with the multiprocessing."""
from __future__ import division, unicode_literals

import logging
import os
import threading

LOGGER = logging.getLogger(__name__)


# On Windows, some functions and classes are monkey-patched in order to keep it
# compatible with older Python 2.7 interpreters. Particularly, issues have been
# encoutered with the 2.7.9 interpreter under Windows.
if os.name == "nt":

    class MockValue(object):
        """A mock object for multiprocessing Value."""

        def __init__(self, value):
            """Constructor.

            :param value: initializing value
            :type value: `int` or `float`
            """
            self.value = value
            self.lock = threading.RLock()

        def get_lock(self):
            """Get the Lock."""
            return self.lock

    def new_value(new_type, initial_value):
        """Monkey patch for the Value function.

        :param new_type: type of the Value, currently 'i' or 'd' are supported
        :type new_type: `str`
        :param initial_value: initial value of the Value object
        :type initial_value: `int` or `float`
        """
        if new_type == "i":
            initial_value = int(initial_value)
        elif new_type == "d":
            initial_value = float(initial_value)
        else:
            raise ValueError("%s is not a supported type" % new_type)

        return MockValue(initial_value)

    class NewManager(object):
        """Monkey patch for the Manager class."""

        @staticmethod
        def dict():
            """Monkey patch the dict() Manager object."""
            return dict()

    RLock = threading.RLock
    Manager = NewManager
    Value = new_value
else:
    from multiprocessing import Manager, RLock, Value  # noqa: F401



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A factory to create regression models.

This module contains a factory to instantiate a :class:`.MLRegressionAlgo` from its
class name. It also provides the available regression models and allows to test if a
regression model type is available.
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.factory import Factory
from gemseo.mlearning.core.factory import MLAlgoFactory
from gemseo.mlearning.regression.regression import MLRegressionAlgo

LOGGER = logging.getLogger(__name__)


class RegressionModelFactory(MLAlgoFactory):
    """This factory instantiates a :class:`.MLRegressionAlgo` from its class name.

    The class can be either internal to |g| or external. In this second case, it can be
    either implemented in a module referenced in the "GEMSEO_PATH" or in a module The
    class can be either internal to |g| or external. In the second case, it can be
    either implemented in a module referenced in the GEMSEO_PATH environment variable or
    in a module starting with "gemseo_" and referenced in the PYTHONPATH environment
    variable.
    """

    def __init__(self):  # type: (...) ->  None
        super(RegressionModelFactory, self).__init__()
        self.factory = Factory(MLRegressionAlgo, ("gemseo.mlearning.regression",))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#          Arthur Piat: greatly improve the N2 layout
"""Graph-based analysis of the weak and strong couplings between several disciplines."""

from __future__ import unicode_literals

import itertools
import logging
from typing import Dict, List, Sequence, Set, Tuple, Union

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.text import Text
from numpy import array
from pylab import gca

from gemseo.core.dependency_graph import DependencyGraph
from gemseo.core.discipline import MDODiscipline
from gemseo.utils.n2d3.n2_html import N2HTML
from gemseo.utils.py23_compat import Path, string_types

LOGGER = logging.getLogger(__name__)

NodeType = Tuple[List[str], List[str]]
EdgesType = Dict[int, Dict[int, List[str]]]
GraphType = Dict[int, Set[int]]
ComponentType = Tuple[int]


class MDOCouplingStructure(object):
    """Structure of the couplings between several disciplines.

    The methods of this class include the computation of weak, strong or all couplings.

    Attributes:
        disciplines (Sequence[MDODiscipline]): The disciplines.
        graph (DependencyGraph): The directed graph of the disciplines.
        sequence (List[List[Tuple[MDODiscipline]]]): The sequence of execution
            of the disciplines.
    """

    def __init__(
        self,
        disciplines,  # type: Sequence[MDODiscipline]
    ):  # type: (...) -> None
        """
        Args:
            disciplines: The disciplines that possibly exchange coupling variables.
        """
        self.disciplines = disciplines
        self.graph = DependencyGraph(disciplines)
        self.sequence = self.graph.get_execution_sequence()

    @staticmethod
    def is_self_coupled(
        discipline,  # type: MDODiscipline
    ):  # type: (...) -> bool
        """Test if the discipline is self-coupled.

        Self-coupling means that one of its outputs is also an input.

        Args:
            discipline: The discipline.

        Returns:
            Whether the discipline is self-coupled.
        """
        return (
            len(
                set(discipline.get_input_data_names())
                & set(discipline.get_output_data_names())
            )
            > 0
        )

    # methods that determine strong/weak/all couplings
    def strongly_coupled_disciplines(
        self,
        add_self_coupled=True,  # type: bool
        by_group=False,  # type: bool
    ):  # type: (...) -> Union[List[MDODiscipline],List[List[MDODiscipline]]]
        """Determines the strongly coupled disciplines, that is the disciplines that
        occur in (possibly different) MDAs.

        Args:
            add_self_coupled: if True, adds the disciplines that are self-coupled
                to the list of strongly coupled disciplines
            by_group: if True, returns a list of list of strongly coupled diciplines
                where the sublists contains the groups of disciplines that
                are strongly coupled together.
                if False, returns a single list

        Returns:
            The coupled disciplines list or list of list
        """
        strong_disciplines = []
        for parallel_tasks in self.sequence:
            for component in parallel_tasks:
                # find MDAs
                if len(component) > 1:
                    if by_group:
                        strong_disciplines.append(component)
                    else:
                        strong_disciplines += component
                elif add_self_coupled:
                    for discipline in component:
                        if self.is_self_coupled(discipline):
                            if by_group:
                                strong_disciplines.append([discipline])
                            else:
                                strong_disciplines.append(discipline)
        return strong_disciplines

    def weakly_coupled_disciplines(self):  # type: (...) -> List[MDODiscipline]
        """Determine the weakly coupled disciplines.

        These are the disciplines that do not occur in MDAs.

        Returns:
            The weakly coupled disciplines.
        """
        weak_disciplines = []
        for parallel_tasks in self.sequence:
            for component in parallel_tasks:
                # find single disciplines
                if len(component) == 1 and not self.is_self_coupled(component[0]):
                    weak_disciplines.append(component[0])
        return weak_disciplines

    def strong_couplings(self):  # type: (...) -> List[str]
        """Determine the strong couplings.

        These are the outputs of the strongly coupled disciplines
        that are also inputs of the strongly coupled disciplines.

        Returns:
            The names of the strongly coupling variables.
        """
        strong_disciplines = self.strongly_coupled_disciplines(
            add_self_coupled=True, by_group=True
        )
        # determine strong couplings = the outputs of the strongly coupled
        # disciplines that are inputs of any other discipline
        strong_couplings = set()

        for group in strong_disciplines:
            inputs = itertools.chain(*[disc.get_input_data_names() for disc in group])
            outputs = itertools.chain(*[disc.get_output_data_names() for disc in group])
            strong_couplings.update(set(inputs) & set(outputs))

        return sorted(list(strong_couplings))

    def weak_couplings(self):  # type: (...) -> List[str]
        """Determine the weak couplings.

        These are the outputs of the weakly coupled disciplines.

        Returns:
            The names of the weakly coupling variables.
        """
        weak_disciplines = self.weakly_coupled_disciplines()
        # determine strong couplings = the outputs of the weakly coupled
        # disciplines that are inputs of any other discipline
        weak_couplings = set()
        for weak_discipline in weak_disciplines:
            weak_couplings.update(weak_discipline.get_output_data_names())
        return sorted(list(weak_couplings))

    def input_couplings(
        self,
        discipline,  # type: MDODiscipline
    ):  # type: (...) -> List[str]
        """Compute all the input coupling variables of a discipline.

        Args:
            discipline: The discipline.

        Returns:
            The names of the input coupling variables.
        """
        input_names = discipline.get_input_data_names()
        strong_couplings = self.strong_couplings()
        return sorted([name for name in input_names if name in strong_couplings])

    def get_all_couplings(self):  # type: (...) -> List[str]
        """Compute all the weak and strong coupling variables.

        Returns:
            The names of the coupling variables.
        """
        inputs = []
        outputs = []
        for discipline in self.disciplines:
            inputs += discipline.get_input_data_names()
            outputs += discipline.get_output_data_names()
        return sorted(list(set(inputs) & set(outputs)))

    def output_couplings(
        self,
        discipline,  # type: MDODiscipline
        strong=True,  # type: bool
    ):  # type: (...) -> List[str]
        """Compute the output coupling variables of a discipline, either strong or weak.

        Args:
            discipline: The discipline.
            strong: If True, consider the strong couplings. Otherwise, the weak ones.

        Returns:
            The names of the output coupling variables.
        """
        output_names = discipline.get_output_data_names()
        if strong:
            couplings = self.strong_couplings()
        else:
            couplings = self.get_all_couplings()

        return sorted([name for name in output_names if name in couplings])

    def find_discipline(
        self,
        output,  # type: str
    ):  # type: (...) -> MDODiscipline
        """Find which discipline produces a given output.

        Args:
            output: The name of an output.

        Returns:
            The discipline producing this output, if it exists.

        Raises:
            TypeError: If the name of the output is not a string.
            ValueError: If the output is not an output of the discipline.
        """
        if not isinstance(output, string_types):
            raise TypeError("Output shall be a string")
        for discipline in self.disciplines:
            if discipline.is_output_existing(output):
                return discipline
        raise ValueError("{} is not the output of a discipline".format(output))

    def plot_n2_chart(
        self,
        file_path="n2.pdf",  # type: str
        show_data_names=True,  # type:bool
        save=True,  # type:bool
        show=False,  # type: bool
        figsize=(15, 10),  # type: Tuple[int]
        open_browser=False,  # type:bool
    ):  # type: (...) -> None
        """Generate a N2 plot for the disciplines.

        Args:
            file_path: The name of the file path of the figure.
            show_data_names: If ``True``, show the names of the coupling data ;
                otherwise,
                circles are drawn,
                which size depend on the number of coupling names.
            save: If True, save the figure to file_path.
            show: If True, show the plot.
            figsize: The width and height of the figure.
            open_browser: If True, open a browser and display an interactive N2 chart.

        Raises:
            ValueError: If there is only 1 discipline.
        """
        output_directory_path = Path(file_path).parent
        html_file_name = "n2.html"
        html_file_path = output_directory_path / html_file_name
        N2HTML(html_file_path, open_browser).from_graph(self.graph)

        fig = plt.figure(figsize=figsize)
        plt.grid(True)
        axe = gca()
        axe.grid(True, linestyle="-", color="black", lw=1)
        n_disciplines = len(self.disciplines)
        if n_disciplines == 1:
            raise ValueError("N2 Diagrams can be generated for at least 2 disciplines.")
        ax_ticks = list(range(n_disciplines + 1))
        axe.xaxis.set_ticks(ax_ticks)
        axe.yaxis.set_ticks(ax_ticks)
        axe.xaxis.set_ticklabels([])
        axe.yaxis.set_ticklabels([])
        axe.set(xlim=(0, ax_ticks[-1]), ylim=(0, ax_ticks[-1]))
        axe.tick_params(axis="x", direction="in")
        axe.tick_params(axis="y", direction="in")
        fig.tight_layout()

        for discipline_index, discipline in enumerate(self.disciplines):
            x_1 = discipline_index
            x_2 = discipline_index + 1
            y_1 = n_disciplines - discipline_index
            y_2 = n_disciplines - discipline_index - 1
            plt.fill(
                [x_1, x_1, x_2, x_2], [y_1, y_2, y_2, y_1], "limegreen", alpha=0.45
            )
            discipline_name = plt.text(
                discipline_index + 0.5,
                n_disciplines - discipline_index - 0.5,
                discipline.name,
                verticalalignment="center",
                horizontalalignment="center",
            )
            self._check_size_text(discipline_name, fig, n_disciplines)

        couplings = self.graph.get_disciplines_couplings()

        max_coupling_length = array([len(coupling[2]) for coupling in couplings]).max()

        for source, destination, variables in couplings:
            source_position = self.disciplines.index(source)
            destination_position = self.disciplines.index(destination)
            if show_data_names:
                variables_names = plt.text(
                    destination_position + 0.5,
                    n_disciplines - source_position - 0.5,
                    "\n".join(variables),
                    verticalalignment="center",
                    horizontalalignment="center",
                )
                self._check_size_text(variables_names, fig, n_disciplines)
            else:
                circle = plt.Circle(
                    (0.5 + destination_position, n_disciplines - 0.5 - source_position),
                    len(variables) / (3.0 * max_coupling_length),
                    color="blue",
                )
                axe.add_artist(circle)

        if save:
            plt.savefig(file_path)
        if show:
            plt.show()

    @staticmethod
    def _check_size_text(
        text,  # type: Text
        figure,  # type: Figure
        n_disciplines,  # type: int
    ):  # type: (...)-> None
        """Adapt the size of the figure based on the size of the text to display.

        Args:
            text: The text shown in the N2 matrix.
            figure: The figure of the N2 matrix.
            n_disciplines: The number of disciplines to be visible.
        """
        renderer = figure.canvas.get_renderer()
        bbox = text.get_window_extent(renderer=renderer)
        width = bbox.width
        height = bbox.height
        size_max_box = figure.get_size_inches() * figure.dpi / n_disciplines
        inches = figure.get_size_inches()
        if width > size_max_box[0]:
            width_l = round(width * n_disciplines / figure.dpi) + 1
            figure.set_size_inches(width_l, inches[1])
        if height > size_max_box[1]:
            length_l = round(height * n_disciplines / figure.dpi) + 1
            figure.set_size_inches(inches[0], length_l)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create an uniform distribution from the SciPy library.

This class inherits from :class:`.SPDistribution`.
"""

from __future__ import division, unicode_literals

from gemseo.uncertainty.distributions.scipy.distribution import SPDistribution


class SPUniformDistribution(SPDistribution):
    """Create an uniform distribution.

    Example:
        >>> from gemseo.uncertainty.distributions.scipy.uniform import (
        ...     SPUniformDistribution
        ... )
        >>> distribution = SPUniformDistribution('x', -1, 1)
        >>> print(distribution)
        uniform(lower=-1, upper=1)
    """

    def __init__(
        self,
        variable,  # type: str
        minimum=0.0,  # type: float
        maximum=1.0,  # type: float
        dimension=1,  # type: int
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the uniform random variable.
            minimum: The minimum of the uniform random variable.
            maximum: The maximum of the uniform random variable.
            dimension: The dimension of the uniform random variable.
        """
        parameters = {"loc": minimum, "scale": maximum - minimum}
        standard_parameters = {self._LOWER: minimum, self._UPPER: maximum}
        super(SPUniformDistribution, self).__init__(
            variable, "uniform", parameters, dimension, standard_parameters
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalability study - API
=======================

This API facilitates the use of the :mod:`gemseo.problems.scalable.data_driven.study`
package implementing classes to benchmark MDO formulations
based on scalable disciplines.

:class:`.ScalabilityStudy` class implements the concept of scalability study:

1. By instantiating a :class:`.ScalabilityStudy`, the user defines
   the MDO problem in terms of design parameters, objective function and
   constraints.
2. For each discipline, the user adds a dataset stored
   in a :class:`.Dataset` and select a type of
   :class:`.ScalableModel` to build the :class:`.ScalableDiscipline`
   associated with this discipline.
3. The user adds different optimization strategies, defined in terms
   of both optimization algorithms and MDO formulation.
4. The user adds different scaling strategies, in terms of sizes of
   design parameters, coupling variables and equality and inequality
   constraints. The user can also define a scaling strategies according to
   particular parameters rather than groups of parameters.
5. Lastly, the user executes the :class:`.ScalabilityStudy` and the results
   are written in several files and stored into directories
   in a hierarchical way, where names depends on both MDO formulation,
   scaling strategy and replications when it is necessary. Different kinds
   of files are stored: optimization graphs, dependency matrix plots and
   of course, scalability results by means of a dedicated class:
   :class:`.ScalabilityResult`.
"""
from __future__ import division, unicode_literals

from gemseo.problems.scalable.data_driven.study.post import PostScalabilityStudy
from gemseo.problems.scalable.data_driven.study.process import ScalabilityStudy


def create_scalability_study(
    objective,
    design_variables,
    directory="study",
    prefix="",
    eq_constraints=None,
    ineq_constraints=None,
    maximize_objective=False,
    fill_factor=0.7,
    active_probability=0.1,
    feasibility_level=0.8,
    start_at_equilibrium=True,
    early_stopping=True,
    coupling_variables=None,
):
    """This method creates a :class:`.ScalabilityStudy`. It requires two mandatory
    arguments:

    - the :code:`'objective'` name,
    - the list of :code:`'design_variables'` names.

    Concerning output files, we can specify:

    - the :code:`directory` which is :code:`'study'` by default,
    - the prefix of output file names (default: no prefix).

    Regarding optimization parametrization, we can specify:

    - the list of equality constraints names (:code:`eq_constraints`),
    - the list of inequality constraints names (:code:`ineq_constraints`),
    - the choice of maximizing the objective function
      (:code:`maximize_objective`).

    By default, the objective function is minimized and the MDO problem
    is unconstrained.

    Last but not least, with regard to the scalability methodology,
    we can overwrite:

    - the default fill factor of the input-output dependency matrix
      :code:`ineq_constraints`,
    - the probability to set the inequality constraints as active at
      initial step of the optimization :code:`active_probability`,
    - the offset of satisfaction for inequality constraints
      :code:`feasibility_level`,
    - the use of a preliminary MDA to start at equilibrium
      :code:`start_at_equilibrium`,
    - the post-processing of the optimization database to get results
      earlier than final step :code:`early_stopping`.

    :param str objective: name of the objective
    :param list(str) design_variables: names of the design variables
    :param str directory: working directory of the study. Default: 'study'.
    :param str prefix: prefix for the output filenames. Default: ''.
    :param list(str) eq_constraints: names of the equality constraints.
        Default: None.
    :param list(str) ineq_constraints: names of the inequality constraints
        Default: None.
    :param bool maximize_objective: maximizing objective. Default: False.
    :param float fill_factor: default fill factor of the input-output
        dependency matrix. Default: 0.7.
    :param float active_probability: probability to set the inequality
        constraints as active at initial step of the optimization.
        Default: 0.1
    :param float feasibility_level: offset of satisfaction for inequality
        constraints. Default: 0.8.
    :param bool start_at_equilibrium: start at equilibrium
        using a preliminary MDA. Default: True.
    :param bool early_stopping: post-process the optimization database
        to get results earlier than final step.
    """
    return ScalabilityStudy(
        objective,
        design_variables,
        directory,
        prefix,
        eq_constraints,
        ineq_constraints,
        maximize_objective,
        fill_factor,
        active_probability,
        feasibility_level,
        start_at_equilibrium,
        early_stopping,
        coupling_variables,
    )


def plot_scalability_results(study_directory):
    """This method plots the set of :class:`ScalabilityResult` generated by a
    :class:`.ScalabilityStudy` and located in the directory created by this study.

    :param str study_directory: directory of the scalability study.
    "
    """
    return PostScalabilityStudy(study_directory)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains the baseclass for regression algorithms.

The :mod:`~gemseo.mlearning.regression.regression` module
implements regression algorithms,
where the goal is to find relationships
between continuous input and output variables.
After being fitted to a learning set,
the regression algorithms can predict output values of new input data.

A regression algorithm consists of identifying a function
:math:`f: \\mathbb{R}^{n_{\\textrm{inputs}}} \\to
\\mathbb{R}^{n_{\\textrm{outputs}}}`.
Given an input point
:math:`x \\in \\mathbb{R}^{n_{\\textrm{inputs}}}`,
the predict method of the regression algorithm will return
the output point :math:`y = f(x) \\in \\mathbb{R}^{n_{\\textrm{outputs}}}`.
See :mod:`~gemseo.mlearning.core.supervised` for more information.

Wherever possible,
the regression algorithms should also be able
to compute the Jacobian matrix of the function it has learned to represent.
Thus,
given an input point :math:`x \\in \\mathbb{R}^{n_{\\textrm{inputs}}}`,
the Jacobian prediction method of the regression algorithm should return the matrix

.. math::

    J_f(x) = \\frac{\\partial f}{\\partial x} =
    \\begin{pmatrix}
    \\frac{\\partial f_1}{\\partial x_1} & \\cdots & \\frac{\\partial f_1}
        {\\partial x_{n_{\\textrm{inputs}}}}\\\\
    \\vdots & \\ddots & \\vdots\\\\
    \\frac{\\partial f_{n_{\\textrm{outputs}}}}{\\partial x_1} & \\cdots &
        \\frac{\\partial f_{n_{\\textrm{outputs}}}}
        {\\partial x_{n_{\\textrm{inputs}}}}
    \\end{pmatrix}
    \\in \\mathbb{R}^{n_{\\textrm{outputs}}\\times n_{\\textrm{inputs}}}.

This concept is implemented through the :class:`.MLRegressionAlgo` class
which inherits from the :class:`.MLSupervisedAlgo` class.
"""
from __future__ import division, unicode_literals

from typing import Callable, Iterable, NoReturn, Optional

from numpy import eye, matmul, ndarray

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, MLAlgoParameterType, TransformerType
from gemseo.mlearning.core.supervised import MLSupervisedAlgo
from gemseo.mlearning.transform.scaler.min_max_scaler import MinMaxScaler
from gemseo.utils.data_conversion import DataConversion


class MLRegressionAlgo(MLSupervisedAlgo):
    """Machine Learning Regression Model Algorithm.

    Inheriting classes shall implement the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLSupervisedAlgo._predict` methods, and
    :meth:`!MLRegressionAlgo._predict_jacobian` method if possible.
    """

    DEFAULT_TRANSFORMER = {
        Dataset.INPUT_GROUP: MinMaxScaler(),
        Dataset.OUTPUT_GROUP: MinMaxScaler(),
    }

    def __init__(
        self,
        data,  # type: Dataset
        transformer=DEFAULT_TRANSFORMER,  # type: TransformerType
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        **parameters  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        super(MLRegressionAlgo, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            **parameters
        )

    class DataFormatters(MLSupervisedAlgo.DataFormatters):
        """Machine learning regression model decorators."""

        @classmethod
        def format_dict_jacobian(
            cls,
            predict_jac,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...) -> Callable[[DataType],DataType]
            """Wrap an array-based function to make it callable with a dictionary of
            NumPy arrays.

            Args:
                predict_jac: The function to be called;
                    it takes a NumPy array in input and returns a NumPy array.

            Returns:
                The wrapped 'predict_jac' function, callable with
                either a NumPy data array
                or a dictionary of numpy data arrays indexed by variables names.
                The return value will have the same type as the input data.
            """

            def wrapper(self, input_data, *args, **kwargs):
                """Evaluate 'predict_jac' with either array or dictionary-based data.

                Firstly,
                the pre-processing stage converts the input data to a NumPy data array,
                if these data are expressed as a dictionary of NumPy data arrays.

                Then,
                the processing evaluates the function 'predict_jac'
                from this NumPy input data array.

                Lastly,
                the post-processing transforms the output data
                to a dictionary of output NumPy data array
                if the input data were passed as a dictionary of NumPy data arrays.

                Args:
                    input_data: The input data.
                    *args: The positional arguments of the function 'predict_jac'.
                    **kwargs: The keyword arguments of the function 'predict_jac'.

                Returns:
                    The output data with the same type as the input one.
                """
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                single_sample = len(input_data.shape) == 1
                jacobians = predict_jac(self, input_data, *args, **kwargs)
                if as_dict:
                    varsizes = self.learning_set.sizes
                    if single_sample:
                        jacobians = DataConversion.jac_2dmat_to_dict(
                            jacobians, self.output_names, self.input_names, varsizes
                        )
                    else:
                        jacobians = DataConversion.jac_3dmat_to_dict(
                            jacobians, self.output_names, self.input_names, varsizes
                        )
                return jacobians

            return wrapper

        @classmethod
        def transform_jacobian(
            cls,
            predict_jac,  # type: Callable[[ndarray],ndarray]
        ):  # type: (...) -> Callable[[ndarray],ndarray]
            """Apply transformation to inputs and inverse transformation to outputs.

            Args:
                predict_jac: The function of interest to be called.

            Returns:
                A function evaluating the function 'predict_jac',
                after transforming its input data
                and/or before transforming its output data.
            """

            def wrapper(self, input_data, *args, **kwargs):
                """Evaluate 'predict_jac' after or before data transformation.

                Firstly,
                the pre-processing stage transforms the input data if required.

                Then,
                the processing evaluates the function 'predict_jac'.

                Lastly,
                the post-processing stage transforms the output data if required.

                Args:
                    input_data: The input data.
                    *args: The positional arguments of the function.
                    **kwargs: The keyword arguments of the function.

                Returns:
                    Either the raw output data of 'predict_jac'
                    or a transformed version according to the requirements.
                """
                inputs = self.learning_set.INPUT_GROUP
                if inputs in self.transformer:
                    jac = self.transformer[inputs].compute_jacobian(input_data)
                    input_data = self.transformer[inputs].transform(input_data)
                else:
                    jac = eye(input_data.shape[1])

                jac = matmul(predict_jac(self, input_data, *args, **kwargs), jac)
                output_data = self.predict_raw(input_data)

                outputs = self.learning_set.OUTPUT_GROUP
                if outputs in self.transformer:
                    jac = matmul(
                        self.transformer[outputs].compute_jacobian_inverse(output_data),
                        jac,
                    )
                return jac

            return wrapper

    def predict_raw(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Predict output data from input data.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The predicted output data with shape (n_samples, n_outputs).
        """
        return self._predict(input_data)

    @DataFormatters.format_dict_jacobian
    @DataFormatters.format_samples
    @DataFormatters.transform_jacobian
    def predict_jacobian(
        self,
        input_data,  # type: DataType
    ):  # type: (...) -> NoReturn
        """Predict the Jacobians of the regression model at input_data.

        The user can specify these input data either as a NumPy array,
        e.g. :code:`array([1., 2., 3.])`
        or as a dictionary,
        e.g.  :code:`{'a': array([1.]), 'b': array([2., 3.])}`.

        If the NumPy arrays are of dimension 2,
        their i-th rows represent the input data of the i-th sample;
        while if the NumPy arrays are of dimension 1,
        there is a single sample.

        The type of the output data and the dimension of the output arrays
        will be consistent
        with the type of the input data and the size of the input arrays.

        Args:
            input_data: The input data.

        Returns:
            The predicted Jacobian data.
        """
        return self._predict_jacobian(input_data)

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> NoReturn
        """Predict the Jacobian matrices of the regression model at input_data.

        Args:
            input_data: The input data with shape (n_samples, n_inputs).

        Returns:
            The predicted Jacobian data with shape (n_samples, n_outputs, n_inputs).
        """
        name = self.__class__.__name__
        raise NotImplementedError("Derivatives are not available for {}".format(name))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class to create a probability distribution from the OpenTURNS library.

The :class:`.OTDistribution` class is a concrete class
inheriting from :class:`.Distribution` which is an abstract one.
OT stands for `OpenTURNS <http://www.openturns.org/>`_
which is the library it relies on.

The :class:`.OTDistribution` of a given uncertain variable is built
from mandatory arguments:

- a variable name,
- a distribution name recognized by OpenTURNS,
- a set of parameters provided as a tuple
  of positional arguments filled in the order
  specified by the OpenTURNS constructor of this distribution.

.. warning::

    The distribution parameters must be provided according to the signature
    of the openTURNS classes. `Access the openTURNS documentation
    <http://openturns.github.io/openturns/latest/user_manual/
    probabilistic_modelling.html>`_.

The constructor has also optional arguments:

- a variable dimension (default: 1),
- a standard representation of these parameters
  (default: use the parameters provided in the tuple),
- a transformation of the variable (default: no transformation),
- lower and upper bounds for truncation (default: no truncation),
- a threshold for the OpenTURNS truncation tool
  (`more details <http://openturns.github.io/openturns/latest/user_manual/
  _generated/openturns.TruncatedDistribution.html>`_).
"""

from __future__ import division, unicode_literals

import logging
from typing import Callable, Iterable, List, Optional

import openturns as ots
from numpy import array, inf, ndarray

from gemseo.uncertainty.distributions.distribution import (
    Distribution,
    ParametersType,
    StandardParametersType,
)
from gemseo.uncertainty.distributions.openturns.composed import OTComposedDistribution
from gemseo.utils.string_tools import MultiLineString

OT_WEBSITE = "http://openturns.github.io/openturns/latest/"
OT_WEBSITE += "user_manual/probabilistic_modelling.html"

LOGGER = logging.getLogger(__name__)


class OTDistribution(Distribution):
    """OpenTURNS probability distribution.

    Create a probability distribution for an uncertain variable
    from its dimension and distribution names and properties.

    Example:
        >>> from gemseo.uncertainty.distributions.openturns.distribution import (
        ...     OTDistribution
        ... )
        >>> distribution = OTDistribution('x', 'Exponential', (3, 2))
        >>> print(distribution)
        Exponential(3, 2)
    """

    _COMPOSED_DISTRIBUTION = OTComposedDistribution

    def __init__(
        self,
        variable,  # type: str
        interfaced_distribution,  # type: str
        parameters,  # type: ParametersType
        dimension=1,  # type: int
        standard_parameters=None,  # type: Optional[StandardParametersType]
        transformation=None,  # type: Optional[str]
        lower_bound=None,  # type: Optional[float]
        upper_bound=None,  # type: Optional[float]
        threshold=0.5,  # type: float
    ):  # noqa: D205,D212,D415
        # type: (...) -> None
        """
        Args:
            variable: The name of the random variable.
            distribution: The name of the probability distribution,
                typically the name of a class wrapped from an external library,
                such as 'Normal' for OpenTURNS or 'norm' for SciPy.
            parameters: The parameters of the probability distribution.
            dimension: The dimension of the random variable.
            standard_parameters: The standard representation
                of the parameters of the probability distribution.
            transformation: A transformation applied
                to the random variable,
                e.g. 'sin(x)'. If None, no transformation.
            lower_bound: A lower bound to truncate the distribution.
                If None, no lower truncation.
            upper_bound: An upper bound to truncate the distribution.
                If None, no upper truncation.
            threshold: A threshold in [0,1].
        """
        super(OTDistribution, self).__init__(
            variable,
            interfaced_distribution,
            parameters,
            dimension,
            standard_parameters,
        )
        self.marginals = self.__create_distributions(
            self.distribution_name,
            self.parameters,
            transformation,
            lower_bound,
            upper_bound,
            threshold,
        )
        self.distribution = ots.ComposedDistribution(self.marginals)
        msg = MultiLineString()
        msg.indent()
        msg.add("Mathematical support: {}", self.support)
        msg.add("Numerical range: {}", self.range)
        msg.add("Transformation: {}", self.transformation)
        LOGGER.info("%s", msg)

    def compute_samples(
        self,
        n_samples=1,  # type: int
    ):  # noqa: D102
        # type: (...) -> ndarray
        sample = array(self.distribution.getSample(n_samples))
        return sample

    def compute_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        return array(
            [
                self.marginals[index].computeCDF(ots.Point([value]))
                for index, value in enumerate(vector)
            ]
        )

    def compute_inverse_cdf(
        self,
        vector,  # type: Iterable[float]
    ):  # noqa: D102
        # type: (...) -> ndarray
        return array(
            [
                self.marginals[index].computeQuantile(value)[0]
                for index, value in enumerate(vector)
            ]
        )

    def _pdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        def pdf(
            point,  # type: float
        ):
            # type: (...) -> float
            """Probability Density Function (PDF).

            Args:
                point: An evaluation point.

            Returns:
                The PDF value at the evaluation point.
            """
            return self.marginals[index].computePDF(point)

        return pdf

    def _cdf(
        self,
        index,  # type: int
    ):  # noqa: D102
        # type: (...) -> Callable
        def cdf(
            level,  # type: float
        ):
            # type: (...) -> float
            """Cumulative Density Function (CDF).

            Args:
                level: A probability level.

            Returns:
                The CDF value for the probability level.
            """
            return self.marginals[index].computeCDF(level)

        return cdf

    @property
    def mean(self):  # noqa: D102
        # type: (...) -> ndarray
        return array(self.distribution.getMean())

    @property
    def standard_deviation(self):  # noqa: D102
        # type: (...) -> ndarray
        return array(self.distribution.getStandardDeviation())

    def __create_distributions(
        self,
        distribution,  # type: str
        parameters,  # type:ParametersType
        transformation,  # type: str
        lower_bound,  # type:float
        upper_bound,  # type:float
        threshold,  # type:float
    ):  # type: (...) -> List[ots.Distribution]
        """Instantiate an OpenTURNS distribution for each random variable component.

        Args:
            distribution: The name of the distribution.
            parameters: The parameters of the distribution.
            transformation: A transformation applied to the random variable,
                e.g. 'sin(x)'. If None, no transformation.
            lower_bound: A lower truncation bound.
                If None, no lower truncation.
            upper_bound: An upper truncation bound.
                If None, no upper truncation.
            threshold: A threshold value in [0,1].

        Returns:
            The marginal OpenTURNS distributions.
        """
        try:
            ot_dist = getattr(ots, distribution)
        except Exception:
            raise ValueError(
                "{} is an unknown OpenTURNS distribution.".format(distribution)
            )
        try:
            ot_dist = [ot_dist(*parameters)] * self.dimension
        except Exception:
            args = ", ".join([str(val) for val in parameters])
            raise ValueError(
                "Arguments are wrong in {}({}). "
                "More details on: {}.".format(distribution, args, OT_WEBSITE)
            )
        self.__set_bounds(ot_dist)
        if transformation is not None:
            ot_dist = self.__transform_marginal_dist(ot_dist, transformation)
        self.__set_bounds(ot_dist)
        if lower_bound is not None or upper_bound is not None:
            ot_dist = self.__truncate_marginal_dist(
                ot_dist, lower_bound, upper_bound, threshold
            )
        self.__set_bounds(ot_dist)
        return ot_dist

    def __transform_marginal_dist(
        self,
        marginals,  # type: Iterable[ots.Distribution],
        transformation,  # type: str
    ):  # type: (...) -> List[ots.Distribution]
        """Apply the standard transformations on the marginals.

        Examples of transformations: -, +, *, **, sin, exp, log, ...

        Args:
            marginals: The marginal distributions.
            transformation: A transformation applied to the random variable,
                e.g. 'sin(x)'. If None, no transformation.

        Returns:
            The transformed marginal distributions.
        """
        variable_name = self.variable_name
        transformation = transformation.replace(" ", "")
        symbolic_function = ots.SymbolicFunction([self.variable_name], [transformation])
        marginals = [
            ots.CompositeDistribution(symbolic_function, marginal)
            for marginal in marginals
        ]
        prev = self.transformation
        transformation = transformation.replace(variable_name, "({})".format(prev))
        self.transformation = transformation
        return marginals

    def __truncate_marginal_dist(
        self,
        distributions,  # type: Iterable[ots.Distribution]
        lower_bound,  # type: float
        upper_bound,  # type: float
        threshold=0.5,  # type: float
    ):  # type: (...) -> List[ots.Distribution]
        """Truncate the distribution of a random variable.

        Args:
            distributions: The distributions.
            lower_bound: A lower bound to truncate the distributions.
                If None, no lower truncation.
            upper_bound: An upper bound to truncate the distributions.
                If None, no upper truncation.
            threshold: A threshold in [0,1].

        Returns:
            The transformed distributions.
        """
        marginals = [
            self.__truncate_distribution(
                dist, index, lower_bound, upper_bound, threshold
            )
            for index, dist in enumerate(distributions)
        ]
        prev = self.transformation
        self.transformation = "Trunc({})".format(prev)
        return marginals

    def __truncate_distribution(
        self,
        distributions,  # type: Iterable[ots.Distribution]
        index,  # type: int
        lower_bound,  # type: float
        upper_bound,  # type: float
        threshold=0.5,  # type: float
    ):  # type: (...) -> List[ots.Distribution]
        """Truncate a distribution with lower bound, upper bound or both.

        Args:
            distributions: The distributions.
            index: A random variable component.
            lower_bound: A lower bound to truncate the distribution.
                If None, no lower truncation.
            upper_bound: An upper bound to truncate the distribution.
                If None, no upper truncation.
            threshold: A threshold in [0,1].

        Returns:
            The truncated distributions.
        """
        if lower_bound is None:
            LOGGER.info(
                "Truncate distribution of component %s above %s.", index, upper_bound
            )
            upper = ots.TruncatedDistribution.UPPER
            current_u_b = self.math_upper_bound[index]
            if upper_bound > current_u_b:
                raise ValueError("u_b is greater than the current upper bound.")
            distributions = ots.TruncatedDistribution(
                distributions, upper_bound, upper, threshold
            )
        elif upper_bound is None:
            LOGGER.info(
                "Truncate distribution of component %s" " below %s.", index, lower_bound
            )
            lower = ots.TruncatedDistribution.LOWER
            current_l_b = self.math_lower_bound[index]
            if lower_bound < current_l_b:
                raise ValueError("l_b is lower " "than the current lower bound.")
            distributions = ots.TruncatedDistribution(
                distributions, lower_bound, lower, threshold
            )
        else:
            LOGGER.info(
                "Truncate distribution of component %s" " below %s and above %s.",
                index,
                lower_bound,
                upper_bound,
            )
            current_l_b = self.math_lower_bound[index]
            current_u_b = self.math_upper_bound[index]
            if lower_bound < current_l_b:
                raise ValueError("l_b is lower " "than the current lower bound.")
            if upper_bound > current_u_b:
                raise ValueError("u_b is greater " "than the current upper bound.")
            distributions = ots.TruncatedDistribution(
                distributions, lower_bound, upper_bound, threshold
            )
        return distributions

    def __set_bounds(
        self, distributions  # type: Iterable[ots.Distribution]
    ):  # type: (...) -> None
        """Set the mathematical and numerical bounds (= support and range).

        Args:
            distributions: The distributions.
        """
        self.math_lower_bound = []
        self.math_upper_bound = []
        self.num_lower_bound = []
        self.num_upper_bound = []
        for distribution in distributions:
            dist_range = distribution.getRange()
            l_b = dist_range.getLowerBound()[0]
            u_b = dist_range.getUpperBound()[0]
            self.num_lower_bound.append(l_b)
            self.num_upper_bound.append(u_b)
            if not dist_range.getFiniteLowerBound()[0]:
                l_b = -inf
            if not dist_range.getFiniteUpperBound()[0]:
                u_b = inf
            self.math_lower_bound.append(l_b)
            self.math_upper_bound.append(u_b)
        self.math_lower_bound = array(self.math_lower_bound)
        self.math_upper_bound = array(self.math_upper_bound)
        self.num_lower_bound = array(self.num_lower_bound)
        self.num_upper_bound = array(self.num_upper_bound)
